/*
   ; Initialize and control one or two pan-tilt cameras
   ; ==================================================

   ;------------------------------------------------------------------------
   ; Author:    Skye Legon  6/2/01   (slegon@k-team.com)
   ; Modifs:
   ;
   ; This program will initialize one or two pan-tilt cameras connected
   ; to a Kameleon controller board.  It auto-detects whether there are
   ; one or two cameras, and then performs a calibration routine to
   ; find the center positions and limits.
   ;
   ; It then provides a simple serial protocol to command the pan and
   ; tilt axes of the camera(s), suitable for manual testing or for use
   ; with a Koala robot.
   ;
   ; NOTE: the current compiler package does not redirect the output of
   ;       "printf" to the COM port, but instead to the SER port.  However
   ;       in mode MMA only messages written to the COM port are passed
   ;       via MMA to the Koala.  To temporarily get around this, this program
   ;       replaces all "printf" commands with a "Send" function that uses
   ;       the COM port.
   ;
   ;------------------------------------------------------------------------
 */

#define TEST_SPEED 		20
#define TORQUE_OFFSET 11
#define TORQUE_LIMIT 	25

#include 	<sys/kos.h>
#include 	<stdlib.h>
#include 	<stdio.h>
#include 	<string.h>

int32 _limits[4] = {0,0,0,0};
int32 _pan[2] = {-1,-1};
int32 _tilt[2] = {-1,-1};

void Send(char *string)
{
	// note that com_reserve_channel() does NOT seem to work.  Instead we
	// use com_send_buffer itself to test whether the port is free or not
	while (com_send_buffer(string, strlen(string)) < 0) tim_switch_fast();
}

int32 sgn(int32 num)
{
	return (num < 0 ? -1 : 1);
}

int32 InitializeMotor(uint8 numMotor)
{
	int32 range, speed;
	uint32 torque;

	mot_config_speed_1m(numMotor,150,4,0);	
	mot_put_sensors_1m(numMotor,0);

	// turn slowly 
	mot_new_speed_1m(numMotor,-TEST_SPEED);
	tim_suspend_task(500);
	// check to see if the motor exists 
	// (if it doesn't move AND has low torque, it isn't there!)
	torque = sens_get_ana_value(TORQUE_OFFSET+numMotor);
	if ((mot_get_position(numMotor) == 0) && (torque < TORQUE_LIMIT)) return -1;

	// loop until we hit the stop pin
	do
	{
		tim_suspend_task(100);
		speed = mot_get_speed(numMotor);
	}
	while (abs(speed) > 0);
	
	mot_stop_1m(numMotor);
	mot_put_sensors_1m(numMotor,0);
	// move in other direction now
	mot_new_speed_1m(numMotor,TEST_SPEED);
	do
	{
		tim_suspend_task(100);
		speed = mot_get_speed(numMotor);
		fflush(stdout);
	}
	while (abs(speed) > 0);

	range = mot_get_position(numMotor);
	mot_stop_1m(numMotor);
	mot_config_speed_1m(numMotor,200,0,0);	
	// go to centre
	mot_new_position_1m(numMotor,(range/2));
	// wait until we've stopped moving (check bit 18 "on-position")
	while ((mot_get_status(numMotor) & (0x01 << 18))==0) tim_switch_fast();
	// set center position to be zero
	mot_put_sensors_1m(numMotor,0);
	mot_new_position_1m(numMotor,0);

	// return limit just a bit inside the stop pins
	return (range/2 - 200);
}

int32 NewMotorPosition(uint8 numMotor, int32 newPosition)
{
	char buffer[80];

	if (numMotor>3) return -1; // invalid motor
	if (_limits[numMotor]==0) return -2; // no motor, or motor not configured
	if (abs(newPosition) > _limits[numMotor]) // out of bounds
	{
		sprintf(buffer, "WARNING: motor position %ld out of bounds...", 
			newPosition);
		Send(buffer);
		newPosition = _limits[numMotor] * sgn(newPosition);
		sprintf(buffer, "set to limit %ld.\r\n", newPosition);
		Send(buffer);
	}

	mot_new_position_1m(numMotor, newPosition);

	return 0;
}

void procInit(void *numMotor)
{
	_limits[(uint8)numMotor] = InitializeMotor((uint8)numMotor);

	exit(0);
}

void PrintHelp(uint8 numMotors)
{
	Send("--------------------------                                \r\n");
	Send("     PAN-TILT module                                      \r\n");
	Send("--------------------------                                \r\n");
	if (numMotors==2)
	{
		Send("P,pos     - move PAN motor (center pos = 0)               \r\n");
		Send("P         - print current PAN position                    \r\n");
		Send("T,pos     - move TILT motor (center pos = 0)              \r\n");
		Send("T         - print current TILT position                   \r\n");
	}
	else
	{
		Send("P,cam,pos - move PAN motor (cam = 0 or 1, center pos = 0) \r\n");
		Send("P,cam     - print current PAN position (cam = 0 or 1)     \r\n");
		Send("T,cam,pos - move TILT motor (cam = 0 or 1, center pos = 0)\r\n");
		Send("T,cam     - print current TILT position (cam = 0 or 1)    \r\n");
	}
	Send("x         - exit to SerCom (38400 baud)                   \r\n");
}

int main()
{
	uint8 i=0, numMotor, numMotors=0;
	uint32 taskID[8];
	int32 parms[20];
	int32 receivedByte;
	char userCommand[80];
	char buffer[80];

	ext_reset();
	mot_reset();
	var_reset();
	sens_reset();
	str_reset();

	// configure all 2 or 4 motors simultaneously by launching parallel tasks
	Send("Configuring motors ...\r\n");
	for (numMotor=0; numMotor<4; numMotor++)
	{
		taskID[numMotor] = 
			install_task_parms("Init\r\n", 800, procInit, (void *)numMotor);
	}

	// now wait until all four tasks have finished
	for (numMotor=0; numMotor<4; numMotor++)
	{
		Send("-----------------------\r\n");
		sprintf(buffer, "Motor %d...\r\n", numMotor);
		Send(buffer);
		while (_limits[numMotor] == 0) tim_switch_fast();

		if (_limits[numMotor] < 0)
		{
			sprintf(buffer, "No motor found on port %d.  Skipping...\r\n", numMotor);
			Send(buffer);
		}
		else
		{
			numMotors++;
			sprintf(buffer, 
				"Motor successfully configured.  Range limit: (+/- %ld) \r\n",
				_limits[numMotor]);
			Send(buffer);
		}
	}

	Send("Finished initialization.\r\n");

	if (numMotors==2) // only one camera
	{
		Send("Found one pan-tilt camera.  Auto-detecting PAN and TILT axes...\r\n");
		for (numMotor=0; numMotor<4; numMotor++)
		{
			if (_limits[numMotor] > 2500)
			{
				sprintf(buffer, "TILT axis is motor %d.\r\n", numMotor);
				Send(buffer);
				_tilt[0] = numMotor;
			}
			else if (_limits[numMotor] > 1500)
			{
				sprintf(buffer, "PAN axis is motor %d.\r\n", numMotor);
				Send(buffer);
				_pan[0] = numMotor;
			}
		}
	}
	else if (numMotors==4) // two cameras
	{
		Send("Found two pan-tilt cameras.  Auto-detecting PAN and TILT axes...\r\n");
		for (numMotor=0; numMotor<4; numMotor++)
		{
			if (_limits[numMotor] > 2500)
			{
				sprintf(buffer, "TILT axis is motor %d.\r\n", numMotor);
				Send(buffer);
				_tilt[numMotor/2] = numMotor;
			}
			else if (_limits[numMotor] > 1500)
			{
				sprintf(buffer, "PAN axis is motor %d.\r\n", numMotor);
				Send(buffer);
				_pan[numMotor/2] = numMotor;
			}
		}
	}
	else
	{
		// if we find 0, 1, or 3 motors signal an error
		sprintf(buffer, 
			"Unrecognized configuration (%d motors).  Exiting.\r\n", numMotors);
		Send(buffer);
		tim_suspend_task(500);
		bios_restart_launch('FUSC', 6); 
	}

	// now start serial command protocol

	PrintHelp(numMotors);

	// clean buffer
	while (com_receive_byte() >= 0) tim_switch_fast();

	for(;;)
	{
		i = 0;
		do
		{
			if ((receivedByte = com_receive_byte()) > 0)
			{
				// restart SerCom 38400 on 'x'
				if ((char)receivedByte == 'x') bios_restart_launch('FUSC', 6); 
				// check for backspace
				else if (((char)receivedByte == '\b')&&(i>0)) i--;
				// accept only "normal" characters
				else if (((char)receivedByte >= ' ')&&((char)receivedByte <= 'z'))
					userCommand[i++] = (char)receivedByte;
			}
			else
				tim_switch_fast();
		}
		while ((char)receivedByte != '\r');

		userCommand[i++] = '\0';

		switch (userCommand[0])
		{
			case 'P':
			{
				if (numMotors==4)
				{
					if(sscanf(userCommand, "P,%ld,%ld", &parms[0], &parms[1]) == 2)
					{
						if ((parms[0]==0)||(parms[0]==1))
						{
							NewMotorPosition(_pan[parms[0]], parms[1]);
							Send("p\r\n");
						}
						else
							Send("Invalid camera (0 or 1).\r\n");
					}
					else if(sscanf(userCommand, "P,%ld", &parms[0]) == 1)
					{
						sprintf(buffer, "p,%ld\r\n", mot_get_position(_pan[parms[0]]));
						Send(buffer);
					}
					else
						Send("Did not understand input.\r\n");
				}
				else
				{
					if(sscanf(userCommand, "P,%ld", &parms[0]) == 1)
					{
						NewMotorPosition(_pan[0], parms[0]);
						Send("p\r\n");
					}
					else
					{
						sprintf(buffer, "p,%ld\r\n", mot_get_position(_pan[0]));
						Send(buffer);
					}
				}

				break;
			}
			case 'T':
			{
				if (numMotors==4)
				{
					if(sscanf(userCommand, "T,%ld,%ld", &parms[0], &parms[1]) == 2)
					{
						if ((parms[0]==0)||(parms[0]==1))
						{
							NewMotorPosition(_tilt[parms[0]], parms[1]);
							Send("t\r\n");
						}
						else
							Send("Invalid camera (0 or 1).\r\n");
					}
					else if(sscanf(userCommand, "T,%ld", &parms[0]) == 1)
					{
						sprintf(buffer, "t,%ld\r\n", mot_get_position(_tilt[parms[0]]));
						Send(buffer);
					}
					else
						Send("Did not understand input.\r\n");
				}
				else
				{
					if(sscanf(userCommand, "T,%ld", &parms[0]) == 1)
					{
						NewMotorPosition(_tilt[0], parms[0]);
						Send("t\r\n");
					}
					else
					{
						sprintf(buffer, "t,%ld\r\n", mot_get_position(_tilt[0]));
						Send(buffer);
					}
				}

				break;
			}
			default:
			{
				PrintHelp(numMotors);
			}
		}
	}

	return 0;
}
